/*
 *  Copyright 2003-2014  Mikhail Kryshen <mikhail@kryshen.net>
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
package com.games1729.shariks;

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.util.*;
import java.io.*;
import java.util.concurrent.locks.LockSupport;

/**
 * Initializes the game.
 */
public class Game implements Runnable {

    public static final String TITLE = "Shariks";
    public static final String VERSION = "0.3.0";

    private static final int DEFAULT_WIDTH = 780;
    private static final int DEFAULT_HEIGHT = 546;

    static final Random random = new Random();

    final Component component;
    final Skin skin;
    final Renderer renderer;
    final BufferStrategy bufferStrategy;

    final Field field;

    private int xOffset = 0;
    private int yOffset = 0;
    private int width = 0;
    private int height = 0;
   
    private static final long ONE_SECOND = 1000000000;
    
    private static final long REPAINT_INTERVAL = ONE_SECOND / 60; // 60 fps
    private static final long LAZY_REPAINT_INTERVAL = ONE_SECOND / 2;
    
    private final Thread renderThread = new Thread(this);

    private volatile boolean suspended = false;
    private volatile boolean lazy = true;
    private volatile boolean update = false;

    public Game(Component main, BufferStrategy bs) {
        component = main;
        bufferStrategy = bs;
        skin = new Skin(this);
        renderer = new Renderer(this);
        field = new Field(this);
        component.addKeyListener(field);

        component.addComponentListener(new ComponentAdapter() {
            @Override
            public void componentResized(ComponentEvent e) {
                updateGeometry();
            }
        });

        updateGeometry();
    }

    private void updateGeometry() {
        if (component instanceof Container) {
            Insets insets = ((Container) component).getInsets();
            xOffset = insets.left;
            yOffset = insets.top;
            width = component.getWidth() - insets.left - insets.right;
            height = component.getHeight() - insets.top - insets.bottom;
        } else {
            xOffset = yOffset = 0;
            width = component.getWidth();
            height = component.getHeight();
        }

        float scale = field.getScale();
        xOffset += (width - field.width * scale) / 2;
        yOffset += (height - field.height * scale) / 2;

        update();
    }

    public void startLoad(final InputStream resourceStream) {
        setStatus(TITLE + " " + VERSION + "\nLoading...");

        (new Thread() {
            public void run() {
                try {
                    skin.load(resourceStream);
                } catch (Exception e) {
                    setStatus("Error loading skin:\n" + e);
                }

                renderer.setMessage(TITLE + " " + VERSION
                        + "\nby Mikhail Kryshen");

                Game.this.displayMenu();
            }
        }).start();
    }

    /**
     * Start the render thread.
     */
    public void start(boolean suspended) {
        this.suspended = suspended;
        renderThread.start();
    }

    /**
     * Stop the render thread. Waits for the thread to die.
     */
    public void stop() throws InterruptedException {
        renderThread.interrupt();
        renderThread.join();
    }

    public void suspend() {
        suspended = true;
        LockSupport.unpark(renderThread);
    }

    public void resume() {
        suspended = false;
        LockSupport.unpark(renderThread);
    }

    public void setLazyRenderMode(boolean lazy) {
        this.lazy = lazy;

        Thread t = renderThread;
        if (t != null) {
            LockSupport.unpark(t);
        }
    }

    public void update() {
        if (!lazy) {
            return;
        }

        update = true;
        LockSupport.unpark(renderThread);
    }

    public void run() {
        long interval = 0;
        long frameTime = System.nanoTime();

        renderLoop:
        while (true) {
            do {
                do {
                    Graphics2D g = (Graphics2D) bufferStrategy.getDrawGraphics();

                    g.translate(xOffset, yOffset);

                    g.addRenderingHints(
                            new RenderingHints(
                                    RenderingHints.KEY_ANTIALIASING,
                                    RenderingHints.VALUE_ANTIALIAS_ON));

                    if (field != null) {
                        field.update(interval);
                    }

                    renderer.render(g);

                    g.dispose();

                } while (bufferStrategy.contentsRestored());

                bufferStrategy.show();

            } while (bufferStrategy.contentsLost());

            while (!suspended && !update) {
                long ri = lazy ? LAZY_REPAINT_INTERVAL : REPAINT_INTERVAL;
                long sleep = frameTime + ri - System.nanoTime();

                if (sleep > 0) {
                    LockSupport.parkNanos(sleep);
                }

                if (Thread.interrupted()) {
                    break renderLoop;
                }

                if (sleep <= 0) {
                    break;
                }
            }

            while (suspended) {
                LockSupport.park();

                if (Thread.interrupted()) {
                    break renderLoop;
                }

                frameTime = System.nanoTime();
            }

            update = false;
            interval = System.nanoTime() - frameTime;
            frameTime += interval;
        }
    }

    public void displayMenu() {
        final String[] items = {"Single Player",
            "Two Players",
            "Three Players",
            "Four Players"};

        class MenuKeyListener extends KeyAdapter {

            private int selected = 0;

            MenuKeyListener() {
                repaintMenu();
            }

            public void keyPressed(KeyEvent e) {
                switch (e.getKeyCode()) {
                    case 38: // Up
                        selected--;
                        break;
                    case 40: // Dow
                        selected++;
                        break;
                    case 27: // Esc
                        if (field != null
                                && field.getGameState() == Field.GAME_MENU) {

                            component.removeKeyListener(this);
                            setStatus(null);

                            field.changeGameState(Field.GAME_RUNNING);
                        }
                        return;
                    case 10: // Enter
                        component.removeKeyListener(this);
                        setStatus(null);

                        Player[] players;

                        switch (selected) {
                            case 0:
                                players = new Player[1];
                                players[0] = new Keyboarder(field, Color.GREEN,
                                        "Green player",
                                        Keyboarder.KEYS_1);
                                field.start(players);
                                break;
                            case 1:
                                players = new Player[2];
                                players[0] = new Keyboarder(field, Color.GREEN,
                                        "Green player",
                                        Keyboarder.KEYS_1);
                                players[1] = new Keyboarder(field, Color.BLUE,
                                        "Blue player",
                                        Keyboarder.KEYS_4);
                                field.start(players);
                                break;
                            case 2:
                                players = new Player[3];
                                players[0] = new Keyboarder(field, Color.GREEN,
                                        "Green player",
                                        Keyboarder.KEYS_1);
                                players[1] = new Keyboarder(field, Color.BLUE,
                                        "Blue player",
                                        Keyboarder.KEYS_2);
                                players[2] = new Keyboarder(field, Color.RED,
                                        "Red player",
                                        Keyboarder.KEYS_4);
                                field.start(players);
                                break;
                            case 3:
                                players = new Player[4];
                                players[0] = new Keyboarder(field, Color.GREEN,
                                        "Green player",
                                        Keyboarder.KEYS_1);
                                players[1] = new Keyboarder(field, Color.BLUE,
                                        "Blue player",
                                        Keyboarder.KEYS_2);
                                players[2] = new Keyboarder(field, Color.RED,
                                        "Red player",
                                        Keyboarder.KEYS_3);
                                players[3] = new Keyboarder(field, Color.BLACK,
                                        "Black player",
                                        Keyboarder.KEYS_4);
                                field.start(players);
                                break;
                        }
                    default:
                        return;
                }

                if (selected < 0) {
                    selected = items.length - 1;
                }
                if (selected >= items.length) {
                    selected = 0;
                }

                repaintMenu();
            }

            private void repaintMenu() {
                StringBuilder sb = new StringBuilder(TITLE + " " + VERSION
                        + "            \n            "
                        + "www.games1729.com\n ");
                for (int i = 0; i < items.length; i++) {
                    sb.append('\n');
                    if (i == selected) {
                        sb.append("> ");
                    }
                    sb.append(items[i]);
                    if (i == selected) {
                        sb.append(" <");
                    }
                }

                setStatus(sb.toString());
            }
        }

        component.addKeyListener(new MenuKeyListener());
    }

    private static void runStandalone() {

        final Frame frame = new Frame(TITLE + " " + VERSION);

        frame.setIgnoreRepaint(true);
        frame.setBackground(Color.BLACK);
        frame.setLayout(null);
        frame.setPreferredSize(new Dimension(DEFAULT_WIDTH, DEFAULT_HEIGHT));
        frame.pack();

        Insets insets = frame.getInsets();
        frame.setPreferredSize(new Dimension(
                DEFAULT_WIDTH + insets.left + insets.right,
                DEFAULT_HEIGHT + insets.top + insets.bottom));
        frame.pack();

        frame.createBufferStrategy(2);
        BufferStrategy bufferStrategy = frame.getBufferStrategy();

        final Game game = new Game(frame, bufferStrategy);

        frame.addWindowListener(new WindowAdapter() {
            @Override
            public void windowClosing(WindowEvent e) {
                if (game.field != null) {
                    game.field.changeGameState(Field.GAME_FINISHED);
                }

                try {
                    game.stop();
                } catch (InterruptedException ex) {
                    System.err.println(ex);
                }

                frame.dispose();
            }
        });

//        GraphicsEnvironment.getLocalGraphicsEnvironment().
//                getDefaultScreenDevice().setFullScreenWindow(frame);
	frame.setVisible(true);
        game.start(false);

        try {
            game.startLoad(new FileInputStream("skins/default.bin"));
        } catch (Exception e) {
            game.setStatus("Error loading skin: " + e);
        }
    }

    public static void main(final String[] args) {
        EventQueue.invokeLater(new Runnable() {
            public void run() {
                runStandalone();
            }
        });
    }

    void setStatus(String text) {
        if (renderer != null) {
            renderer.setMessage(text);
        } else if (text != null) {
            System.err.println(text);
        }
    }

    public int getWidth() {
        return width;
    }

    public int getHeight() {
        return height;
    }
}
